#!/usr/bin/env python3
# Copyright (C) 2022 Checkmk GmbH - License: GNU General Public License v2
# This file is part of Checkmk (https://checkmk.com). It is subject to the terms and
# conditions defined in the file COPYING, which is part of this source code package.

#
# file path must be exactly as the file it shadows :-(
# local/lip/python3/cmk/plugins/cisco/rulesets


from collections.abc import Iterable, Sequence

from cmk.rulesets.v1 import Help, Label, Message, Title
from cmk.rulesets.v1.form_specs import (
    # BooleanChoice,
    DefaultValue,
    DictElement,
    Dictionary,
    FixedValue,
    Integer,
    List,
    MultipleChoice,
    MultipleChoiceElement,
    Password,
    Proxy,
    String,
    migrate_to_password,
    migrate_to_proxy,
)

from cmk.rulesets.v1.form_specs.validators import ValidationError, NumberInRange
from cmk.rulesets.v1.rule_specs import SpecialAgent, Topic
from cmk_addons.plugins.meraki.lib.utils import (
    _SEC_CACHE_APPLIANCE_PERFORMANCE,
    _SEC_CACHE_APPLIANCE_UPLINKS_USAGE,
    _SEC_CACHE_APPLIANCE_UPLINKS,
    _SEC_CACHE_APPLIANCE_VPNS,
    _SEC_CACHE_CELLULAR_UPLINKS,
    _SEC_CACHE_DEVICE_INFO,
    _SEC_CACHE_DEVICE_STATUSES,
    _SEC_CACHE_DEVICE_UPLINKS_INFO,
    _SEC_CACHE_LICENSES_OVERVIEW,
    _SEC_CACHE_NETWORKS,
    _SEC_CACHE_ORG_API_REQUESTS,
    _SEC_CACHE_ORG_SWITCH_PORTS_STATUSES,
    _SEC_CACHE_ORGANISATIONS,
    _SEC_CACHE_SENSOR_READINGS,
    _SEC_CACHE_SWITCH_PORTS_STATUSES,
    _SEC_CACHE_WIRELESS_DEVICE_STATUS,
    _SEC_CACHE_WIRELESS_ETHERNET_STATUSES,
)

_SEC_NAME_APPLIANCE_UPLINKS = "appliance_uplinks"
_SEC_NAME_APPLIANCE_UPLINKS_USAGE = "appliance_uplinks_usage"
_SEC_NAME_APPLIANCE_VPNS = "appliance_vpns"
_SEC_NAME_APPLIANCE_PERFORMANCE = "appliance_performance"
_SEC_NAME_CELLULAR_UPLINKS = "cellular_uplinks"
_SEC_NAME_DEVICE_INFO = "device_info"
_SEC_NAME_DEVICE_STATUSES = "device_status"
_SEC_NAME_DEVICE_UPLINKS_INFO = "device_uplinks_info"
_SEC_NAME_LICENSES_OVERVIEW = "licenses_overview"
_SEC_NAME_NETWORKS = "networks"
_SEC_NAME_ORGANISATIONS = "organisations"
_SEC_NAME_ORG_API_REQUESTS = "api_requests_by_organization"
_SEC_NAME_SENSOR_READINGS = "sensor_readings"
_SEC_NAME_SWITCH_PORTS_STATUSES = "switch_ports_statuses"
_SEC_NAME_WIRELESS_DEVICE_STATUS = "wireless_device_status"
_SEC_NAME_WIRELESS_ETHERNET_STATUSES = "wireless_ethernet_statuses"
_SEC_NAME_ORG_SWITCH_PORTS_STATUSES = "org_switch_ports_statuses"

_SEC_TITLE_DEVICE_INFO = 'Device info (Organization)'
_SEC_TITLE_NETWORKS = 'Network info (Organization)'
_SEC_TITLE_ORGANISATIONS = 'Organization (Agent)'
_SEC_TITLE_ORG_API_REQUESTS = 'API request (Organizaion)'
_SEC_TITLE_APPLIANCE_UPLINKS = 'Appliances uplinks (Organizaion)'
_SEC_TITLE_APPLIANCE_UPLINKS_USAGE = 'Appliances uplinks usage (Organizaion)'
_SEC_TITLE_APPLIANCE_VPNS = 'Appliances VPNs (Organizaion)'
_SEC_TITLE_APPLIANCE_PERFORMANCE = 'Appliances Utilization (Device)'
_SEC_TITLE_CELLULAR_UPLINKS = 'Cellular devices uplinks (Organizaion)'
_SEC_TITLE_DEVICE_STATUSES = 'Devices status (Organizaion)'
_SEC_TITLE_DEVICE_UPLINKS_INFO = 'Devices uplink info (Organizaion)'
_SEC_TITLE_LICENSES_OVERVIEW = 'Licenses overview (Organizaion)'
_SEC_TITLE_SENSOR_READINGS = 'Sensors readings (Organizaion)'
_SEC_TITLE_SWITCH_PORTS_STATUSES = 'Switch ports status (Device)'
_SEC_TITLE_WIRELESS_ETHERNET_STATUSES = 'Wireless devices ethernet status (Organizaion)'
_SEC_TITLE_WIRELESS_DEVICE_STATUS = 'Wireless devices SSIDs status (Device)'
_SEC_TITLE_ORG_SWITCH_PORTS_STATUSES = 'Switch port status (Organizaion/Early Access)'


class DuplicateInList:  # pylint: disable=too-few-public-methods
    """ Custom validator that ensures the validated list has no duplicate entries. """

    def __init__(
            self,
    ) -> None:
        pass

    @staticmethod
    def _get_default_errmsg(_duplicates: Sequence) -> Message:
        return Message(f"Duplicate element in list. Duplicate elements: {', '.join(_duplicates)}")

    def __call__(self, value: List[str] | None) -> None:
        if not isinstance(value, list):
            return
        _duplicates = [value[i] for i, x in enumerate(value) if value.count(x) > 1]
        _duplicates = list(set(_duplicates))
        if _duplicates:
            raise ValidationError(message=self._get_default_errmsg(_duplicates))


def _migrate_to_valid_ident(value: object) -> Sequence[str]:
    if not isinstance(value, Iterable):
        raise ValueError("Invalid value {value} for sections")

    name_mapping = {
        "licenses-overview": "licenses_overview",
        "device-statuses": "device_statuses",
        "sensor-readings": "sensor_readings",
    }

    # return [name_mapping.get(s, s) for s in value]
    return [s.replace('-', '_') for s in value]


def _form_special_agent_cisco_meraki() -> Dictionary:
    return Dictionary(
        title=Title("Cisco Meraki"),
        elements={
            "api_key": DictElement(
                parameter_form=Password(
                    title=Title("API Key"),
                    migrate=migrate_to_password
                ),
                required=True,
            ),
            "proxy": DictElement(
                parameter_form=Proxy(
                    migrate=migrate_to_proxy
                )
            ),
            "no_cache": DictElement(
                parameter_form=FixedValue(  # BooleanChoice needs 2 clicks :-(
                    title=Title("Disable Cache"),
                    help_text=Help(
                        "Never use cached information. By default the agent will cache received "
                        "data to avoid API limits and speed up the data retrievel."
                    ),
                    label=Label("API Cache is disabled"),
                    value=True,
                )
            ),
            'org_id_as_prefix': DictElement(
                parameter_form=FixedValue(
                    value=True,
                    title=Title('Uese organisation ID as host prefix'),
                    label=Label("The Organization-id will be used as host name prefix"),
                    help_text=Help(
                        'The organisation ID will be used as prefix for the hostname (separated by a "\'"). Use '
                        'this option together with a "Hostname translation for piggybacked hosts" to add a '
                        'organisation prefix to the hosts from the Cisco Meraki cloud to avoid conflicting '
                        'hostnames. You can also use this option along with the "Dynamic host management" to '
                        'sort the host in organisation specific folders.'
                    )
                )),
            "excluded_sections": DictElement(
                    parameter_form=MultipleChoice(
                        title=Title("Exclude Sections"),
                        elements=[
                            MultipleChoiceElement(name=_SEC_NAME_ORG_API_REQUESTS,
                                                title=Title(_SEC_TITLE_ORG_API_REQUESTS)),
                            MultipleChoiceElement(name=_SEC_NAME_APPLIANCE_UPLINKS,
                                                title=Title(_SEC_TITLE_APPLIANCE_UPLINKS)),
                            MultipleChoiceElement(name=_SEC_NAME_APPLIANCE_UPLINKS_USAGE,
                                                title=Title(_SEC_TITLE_APPLIANCE_UPLINKS_USAGE)),
                            MultipleChoiceElement(name=_SEC_NAME_APPLIANCE_VPNS, title=Title(_SEC_TITLE_APPLIANCE_VPNS)),
                            MultipleChoiceElement(name=_SEC_NAME_APPLIANCE_PERFORMANCE,
                                                title=Title(_SEC_TITLE_APPLIANCE_PERFORMANCE)),
                            MultipleChoiceElement(name=_SEC_NAME_CELLULAR_UPLINKS,
                                                title=Title(_SEC_TITLE_CELLULAR_UPLINKS)),
                            MultipleChoiceElement(name=_SEC_NAME_DEVICE_STATUSES, title=Title(_SEC_TITLE_DEVICE_STATUSES)),
                            MultipleChoiceElement(name=_SEC_NAME_DEVICE_UPLINKS_INFO,
                                                title=Title(_SEC_TITLE_DEVICE_UPLINKS_INFO)),
                            MultipleChoiceElement(name=_SEC_NAME_LICENSES_OVERVIEW,
                                                title=Title(_SEC_TITLE_LICENSES_OVERVIEW)),
                            MultipleChoiceElement(name=_SEC_NAME_SENSOR_READINGS, title=Title(_SEC_TITLE_SENSOR_READINGS)),
                            MultipleChoiceElement(name=_SEC_NAME_SWITCH_PORTS_STATUSES,
                                                title=Title(_SEC_TITLE_SWITCH_PORTS_STATUSES)),
                            MultipleChoiceElement(name=_SEC_NAME_WIRELESS_ETHERNET_STATUSES,
                                                title=Title(_SEC_TITLE_WIRELESS_ETHERNET_STATUSES)),
                            MultipleChoiceElement(name=_SEC_NAME_WIRELESS_DEVICE_STATUS,
                                                title=Title(_SEC_TITLE_WIRELESS_DEVICE_STATUS)),
                            MultipleChoiceElement(name=_SEC_NAME_ORG_SWITCH_PORTS_STATUSES,
                                                title=Title(_SEC_TITLE_ORG_SWITCH_PORTS_STATUSES)),
                        ],
                        prefill=DefaultValue([
                            _SEC_NAME_APPLIANCE_PERFORMANCE,
                            _SEC_NAME_SWITCH_PORTS_STATUSES,
                            _SEC_NAME_WIRELESS_DEVICE_STATUS,
                            _SEC_NAME_ORG_SWITCH_PORTS_STATUSES,
                        ]),
                        # migrate=_migrate_to_valid_ident,
                    ),
                    required=True,
                ),
            "orgs": DictElement(
                parameter_form=List(
                    element_template=String(macro_support=True), title=Title("Organizations"),
                    custom_validate=(DuplicateInList(),),
                ),
            ),
            "cache_per_section": DictElement(
                parameter_form=Dictionary(
                    title=Title("Set Cache time per section"),
                    elements={
                        sec_name: DictElement(
                            parameter_form=Integer(
                                title=Title(sec_title),
                                prefill=DefaultValue(sec_cache),
                                unit_symbol="minutes",
                                custom_validate=(NumberInRange(min_value=0),)
                            )
                        ) for sec_name, sec_title, sec_cache in [
                            (_SEC_NAME_APPLIANCE_PERFORMANCE, _SEC_TITLE_APPLIANCE_PERFORMANCE, _SEC_CACHE_APPLIANCE_PERFORMANCE),
                            (_SEC_NAME_APPLIANCE_UPLINKS_USAGE, _SEC_TITLE_APPLIANCE_UPLINKS_USAGE, _SEC_CACHE_APPLIANCE_UPLINKS_USAGE),
                            (_SEC_NAME_APPLIANCE_UPLINKS, _SEC_TITLE_APPLIANCE_UPLINKS, _SEC_CACHE_APPLIANCE_UPLINKS),
                            (_SEC_NAME_APPLIANCE_VPNS, _SEC_TITLE_APPLIANCE_VPNS, _SEC_CACHE_APPLIANCE_VPNS),
                            (_SEC_NAME_CELLULAR_UPLINKS, _SEC_TITLE_CELLULAR_UPLINKS, _SEC_CACHE_CELLULAR_UPLINKS),
                            (_SEC_NAME_DEVICE_INFO, _SEC_TITLE_DEVICE_INFO, _SEC_CACHE_DEVICE_INFO),
                            (_SEC_NAME_DEVICE_STATUSES, _SEC_TITLE_DEVICE_STATUSES, _SEC_CACHE_DEVICE_STATUSES),
                            (_SEC_NAME_DEVICE_UPLINKS_INFO, _SEC_TITLE_DEVICE_UPLINKS_INFO, _SEC_CACHE_DEVICE_UPLINKS_INFO),
                            (_SEC_NAME_LICENSES_OVERVIEW, _SEC_TITLE_LICENSES_OVERVIEW, _SEC_CACHE_LICENSES_OVERVIEW),
                            (_SEC_NAME_NETWORKS, _SEC_TITLE_NETWORKS, _SEC_CACHE_NETWORKS),
                            (_SEC_NAME_ORG_API_REQUESTS, _SEC_TITLE_ORG_API_REQUESTS, _SEC_CACHE_ORG_API_REQUESTS),
                            (_SEC_NAME_ORG_SWITCH_PORTS_STATUSES, _SEC_TITLE_ORG_SWITCH_PORTS_STATUSES, _SEC_CACHE_ORG_SWITCH_PORTS_STATUSES),
                            (_SEC_NAME_ORGANISATIONS, _SEC_TITLE_ORGANISATIONS, _SEC_CACHE_ORGANISATIONS),
                            (_SEC_NAME_SENSOR_READINGS, _SEC_TITLE_SENSOR_READINGS, _SEC_CACHE_SENSOR_READINGS),
                            (_SEC_NAME_SWITCH_PORTS_STATUSES, _SEC_TITLE_SWITCH_PORTS_STATUSES, _SEC_CACHE_SWITCH_PORTS_STATUSES),
                            (_SEC_NAME_WIRELESS_DEVICE_STATUS, _SEC_TITLE_WIRELESS_DEVICE_STATUS, _SEC_CACHE_WIRELESS_DEVICE_STATUS),
                            (_SEC_NAME_WIRELESS_ETHERNET_STATUSES, _SEC_TITLE_WIRELESS_ETHERNET_STATUSES, _SEC_CACHE_WIRELESS_ETHERNET_STATUSES),
                        ]
                    }
                )
            ),
            "sections": DictElement(
                parameter_form=MultipleChoice(
                    title=Title("Sections"),
                    elements=[
                        MultipleChoiceElement(
                            name="licenses_overview", title=Title("Organization licenses overview")
                        ),
                        MultipleChoiceElement(
                            name="device_statuses", title=Title("Organization device statuses")
                        ),
                        MultipleChoiceElement(
                            name="sensor_readings", title=Title("Organization sensor readings")
                        ),
                    ],
                    # migrate=_migrate_to_valid_ident,
                ),
                render_only=True,
            ),
        },
        # prefill=DefaultValue('excluded_sections'),
    )


rule_spec_cisco_meraki = SpecialAgent(
    name="cisco_meraki",
    title=Title("Cisco Meraki"),
    topic=Topic.APPLICATIONS,
    parameter_form=_form_special_agent_cisco_meraki,
)
